use crate::api;
use crate::core;
use crate::repositories;
use crate::synchronizations;
use beteran_common_rust as bcr;
use beteran_protobuf_rust as bpr;
use diesel::{
  r2d2::{ConnectionManager, Pool},
  PgConnection,
};
use prost::Message;
use std::str::FromStr;

pub struct EventHandler {
  connection_broker: nats::asynk::Connection,
  queue_broker: String,
  pool: Pool<ConnectionManager<PgConnection>>,
  api_config: core::config::ApiConfig,
  member_repository: repositories::member::repository::Repository,
  member_account_api: api::member_account::api::Api,
  member_account_synchronizer: synchronizations::member_account::synchronizer::Synchronizer,
}

impl std::fmt::Debug for EventHandler {
  fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
    f.debug_struct("EventHandler of members").finish()
  }
}

impl EventHandler {
  ///
  pub fn new(
    connection_broker: nats::asynk::Connection,
    queue_broker: String,
    pool: Pool<ConnectionManager<PgConnection>>,
    member_account_synchronizer: synchronizations::member_account::synchronizer::Synchronizer,
    api_config: core::config::ApiConfig,
  ) -> EventHandler {
    EventHandler {
      connection_broker,
      queue_broker,
      pool,
      member_account_synchronizer,
      api_config: api_config.clone(),
      member_repository: repositories::member::repository::Repository::new(),
      member_account_api: api::member_account::api::Api::new(api_config.clone()),
    }
  }

  pub async fn subscribe(&self) -> std::result::Result<(), std::boxed::Box<dyn std::error::Error>> {
    futures::try_join!(self.event_after_update_member_bank_deposit_for_state()).map(|_| ())
  }

  async fn event_after_update_member_bank_deposit_for_state(
    &self,
  ) -> Result<(), Box<dyn std::error::Error>> {
    let s = self
      .connection_broker
      .queue_subscribe(
        bpr::ss::member_bank_deposit::EVENT_SUBJECT_AFTER_UPDATE_MEMBER_BANK_DEPOSIT_FOR_STATE,
        self.queue_broker.as_str(),
      )
      .await?;

    while let Some(message) = s.next().await {
      if let Err(e) = async {
        let eve = bpr::ss::member_bank_deposit::AfterUpdateMemberBankDepositForState::decode(
          message.data.as_slice(),
        )
        .map_err(|e| {
          bcr::error::rpc::Error::InvalidRequest(bcr::error::rpc::InvalidRequest {
            message: format!("invalid request: {}", e),
          })
        })?;

        let client = match eve.client {
          Some(c) => c,
          None => {
            return Err(bcr::error::rpc::Error::InvalidParams(
              bcr::error::rpc::InvalidParams {
                message: "invalid client information".to_string(),
                detail: bcr::error::rpc::InvalidParamsDetail {
                  location: "request".to_string(),
                  param: "client".to_string(),
                  value: "".to_string(),
                  error_type: bcr::error::rpc::InvalidParamsType::Required,
                  message: "".to_string(),
                },
              },
            ));
          }
        };
        let event = match eve.event {
          Some(r) => r,
          None => {
            return Err(bcr::error::rpc::Error::InvalidParams(
              bcr::error::rpc::InvalidParams {
                message: "invalid event information".to_string(),
                detail: bcr::error::rpc::InvalidParamsDetail {
                  location: "event".to_string(),
                  param: "event".to_string(),
                  value: "".to_string(),
                  error_type: bcr::error::rpc::InvalidParamsType::Required,
                  message: "".to_string(),
                },
              },
            ));
          }
        };
        let member_bank_deposit = match event.member_bank_deposit {
          Some(c) => c,
          None => {
            return Err(bcr::error::rpc::Error::InvalidParams(
              bcr::error::rpc::InvalidParams {
                message: "invalid event information".to_string(),
                detail: bcr::error::rpc::InvalidParamsDetail {
                  location: "event".to_string(),
                  param: "member_bank_deposit".to_string(),
                  value: "".to_string(),
                  error_type: bcr::error::rpc::InvalidParamsType::Required,
                  message: "".to_string(),
                },
              },
            ));
          }
        };

        let member = match member_bank_deposit.member {
          Some(c) => c,
          None => {
            return Err(bcr::error::rpc::Error::InvalidParams(
              bcr::error::rpc::InvalidParams {
                message: "invalid event information".to_string(),
                detail: bcr::error::rpc::InvalidParamsDetail {
                  location: "event".to_string(),
                  param: "member_bank_deposit.member".to_string(),
                  value: "".to_string(),
                  error_type: bcr::error::rpc::InvalidParamsType::Required,
                  message: "".to_string(),
                },
              },
            ));
          }
        };
        let member_id = uuid::Uuid::from_str(member.id.as_str()).map_err(|e| {
          bcr::error::rpc::Error::InvalidParams(bcr::error::rpc::InvalidParams {
            message: "invalid member.id param".to_string(),
            detail: bcr::error::rpc::InvalidParamsDetail {
              location: "request".to_string(),
              param: "member.id".to_string(),
              value: member.id.clone(),
              error_type: bcr::error::rpc::InvalidParamsType::Required,
              message: e.to_string(),
            },
          })
        })?;

        let conn = self.pool.get().map_err(|e| {
          bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
            code: bpr::protobuf::rpc::Error::SERVER_00,
            message: format!("server {}", e),
            data: None,
          })
        })?;

        match beteran_protobuf_rust::models::member_bank_deposit::MemberBankDepositState::from_i32(
          member_bank_deposit.state,
        ) {
          Some(s) => {
            if s == beteran_protobuf_rust::models::member_bank_deposit::MemberBankDepositState::Complete {
              self.member_account_api.create_deposit(api::member_account::models::CreateDepositRequest{
                username: member.username.clone(),
                cash_type: None,
                amount: member_bank_deposit.amount,
                request_key: None,
              }).await
              .map_err(|e| {
                bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
                  code: bpr::protobuf::rpc::Error::SERVER_00,
                  message: format!("server {:?}", e),
                  data: None,
                })
              })?;

              self.member_account_synchronizer.balance_for_user(member_id, &member.username).await
              .map_err(|e| {
                bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
                  code: bpr::protobuf::rpc::Error::SERVER_00,
                  message: format!("server {:?}", e),
                  data: None,
                })
              })?;
            }
          }
          None => {}
        };

        Ok::<(), bcr::error::rpc::Error>(())
      }
      .await
      {
        println!("error: {}", e);
      }
    }

    Ok(())
  }
}
