use crate::compositions;
use beteran_common_rust as bcr;
use beteran_protobuf_rust as bpr;
use diesel::{
  r2d2::{ConnectionManager, Pool},
  PgConnection,
};
use prost::Message;

///
pub struct Service {
  connection_broker: nats::asynk::Connection,
  queue_broker: String,
  pool: Pool<ConnectionManager<PgConnection>>,
  betting_composition: compositions::betting::composition::Composition,
}

impl std::fmt::Debug for Service {
  fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
    f.debug_struct("Service of api.kgon.identity").finish()
  }
}

impl Service {
  ///
  pub fn new(
    connection_broker: nats::asynk::Connection,
    queue_broker: String,
    pool: Pool<ConnectionManager<PgConnection>>,
  ) -> Service {
    Service {
      connection_broker: connection_broker.clone(),
      queue_broker,
      pool,
      betting_composition: compositions::betting::composition::Composition::new(),
    }
  }

  pub async fn subscribe(&self) -> std::result::Result<(), std::boxed::Box<dyn std::error::Error>> {
    futures::try_join!(
      self.list_betting_history(),
      self.list_history_for_pragmatic(),
      self.list_history_for_evolution(),
      self.statistics(),
    )
    .map(|_| ())
  }

  async fn list_betting_history(&self) -> Result<(), Box<dyn std::error::Error>> {
    let s = self
      .connection_broker
      .queue_subscribe(
        bpr::ss::api::betting::SUBJECT_LIST_BETTING_HISTORY,
        self.queue_broker.as_str(),
      )
      .await?;

    while let Some(message) = s.next().await {
      if let Err(e) = async {
        let req = bpr::ss::api::betting::ListBettingHistoryRequest::decode(message.data.as_slice())
          .map_err(|e| {
            bcr::error::rpc::Error::InvalidRequest(bcr::error::rpc::InvalidRequest {
              message: format!("invalid request: {}", e),
            })
          })?;

        let client = match req.client {
          Some(c) => c,
          None => {
            return Err(bcr::error::rpc::Error::InvalidParams(
              bcr::error::rpc::InvalidParams {
                message: "invalid client information".to_string(),
                detail: bcr::error::rpc::InvalidParamsDetail {
                  location: "request".to_string(),
                  param: "client".to_string(),
                  value: "".to_string(),
                  error_type: bcr::error::rpc::InvalidParamsType::Required,
                  message: "".to_string(),
                },
              },
            ));
          }
        };
        let request = match req.request {
          Some(r) => r,
          None => {
            return Err(bcr::error::rpc::Error::InvalidParams(
              bcr::error::rpc::InvalidParams {
                message: "invalid request information".to_string(),
                detail: bcr::error::rpc::InvalidParamsDetail {
                  location: "request".to_string(),
                  param: "request".to_string(),
                  value: "".to_string(),
                  error_type: bcr::error::rpc::InvalidParamsType::Required,
                  message: "".to_string(),
                },
              },
            ));
          }
        };

        let conn = self.pool.get().map_err(|e| {
          bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
            code: bpr::protobuf::rpc::Error::SERVER_00,
            message: format!("server {}", e),
            data: None,
          })
        })?;

        let find_all = compositions::betting::models::FindAll {
          search: match request.search {
            Some(s) => Some(compositions::betting::models::FindAllSearch {
              member_id: s.member_id,
              vendor_id: s.vendor_id.map(|v| v as i64),
              vendor_name: s.vendor_name,
              vendor_name_like: s.vendor_name_like,
              game_id: s.game_id.map(|v| v as i64),
              game_name: s.game_name,
              game_name_like: s.game_name_like,
              game_category: s.game_category,
              game_category_like: s.game_category_like,
              game_type: s.game_type,
              game_type_like: s.game_type_like,
              currency: s.currency,
              currency_like: s.currency_like,
              key: s.key,
              key_like: s.key_like,
              ref_id: s.ref_id,
              ref_id_like: s.ref_id_like,
              o_ref_id: s.o_ref_id,
              o_ref_id_like: s.o_ref_id_like,
              group_key: s.group_key,
              group_key_like: s.group_key_like,
              is_bonus: s.is_bonus,
              is_promo: s.is_promo,
              is_jackpot: s.is_jackpot,
              site_username: s.site_username,
              site_username_like: s.site_username_like,
              betting_type: s.betting_type,
              betting_type_like: s.betting_type_like,
              category: s.category,
              category_like: s.category_like,
            }),
            None => None,
          },
          pagination: request
            .pagination
            .as_ref()
            .map(bcr::pagination::Pagination::from),
          sorts: Some(
            request
              .sorts
              .iter()
              .map(bcr::pagination::Sort::from)
              .collect(),
          ),
        };

        let count = self
          .betting_composition
          .select_all_count(&conn, &find_all)
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              code: bpr::protobuf::rpc::Error::SERVER_00,
              message: format!("server {}", e),
              data: None,
            })
          })?;

        let list = self
          .betting_composition
          .select_all(&conn, &find_all)
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              code: bpr::protobuf::rpc::Error::SERVER_00,
              message: format!("server {}", e),
              data: None,
            })
          })?;

        message
          .respond(
            bpr::ss::api::betting::ListBettingHistoryResponse {
              error: None,
              result: Some(
                bpr::ss::api::betting::list_betting_history_response::Result {
                  betting_history: list
                    .iter()
                    .map(bpr::models::api::betting::BettingHistoryModel::from)
                    .collect(),
                },
              ),
            }
            .encode_to_vec(),
          )
          .await
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              code: bpr::protobuf::rpc::Error::SERVER_00,
              message: format!("server {}", e),
              data: None,
            })
          })?;

        Ok::<(), bcr::error::rpc::Error>(())
      }
      .await
      {
        message
          .respond(
            bpr::ss::api::betting::ListBettingHistoryResponse {
              error: Some(bpr::protobuf::rpc::Error::from(e)),
              result: None,
            }
            .encode_to_vec(),
          )
          .await?;
      }
    }

    Ok(())
  }
  async fn list_history_for_pragmatic(&self) -> Result<(), Box<dyn std::error::Error>> {
    Ok(())
  }
  async fn list_history_for_evolution(&self) -> Result<(), Box<dyn std::error::Error>> {
    Ok(())
  }

  async fn statistics(&self) -> Result<(), Box<dyn std::error::Error>> {
    Ok(())
  }
}
