use crate::api;
use crate::core;
use crate::repositories;
use beteran_protobuf_rust as bpr;
use diesel::{
  r2d2::{ConnectionManager, Pool},
  PgConnection,
};
use prost::Message;
use std::net::IpAddr;
use std::net::Ipv4Addr;
use std::str::FromStr;

///
pub struct Synchronizer {
  connection_broker: nats::asynk::Connection,
  pool: Pool<ConnectionManager<PgConnection>>,
  api_config: core::config::ApiConfig,
  synchronization_history_repository: repositories::synchronization_history::repository::Repository,
  member_repository: repositories::member::repository::Repository,
  member_api: api::member::api::Api,
}

impl std::fmt::Debug for Synchronizer {
  fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
    f.debug_struct("Synchronizer of api.kgon.identity").finish()
  }
}

impl std::clone::Clone for Synchronizer {
  fn clone(&self) -> Self {
    Self {
      connection_broker: self.connection_broker.clone(),
      pool: self.pool.clone(),
      api_config: self.api_config.clone(),
      synchronization_history_repository:
        repositories::synchronization_history::repository::Repository::new(),
      member_repository: repositories::member::repository::Repository::new(),
      member_api: api::member::api::Api::new(self.api_config.clone()),
    }
  }
}

impl Synchronizer {
  ///
  pub fn new(
    connection_broker: nats::asynk::Connection,
    pool: Pool<ConnectionManager<PgConnection>>,
    api_config: core::config::ApiConfig,
  ) -> Synchronizer {
    Synchronizer {
      connection_broker,
      pool,
      api_config: api_config.clone(),
      synchronization_history_repository:
        repositories::synchronization_history::repository::Repository::new(),
      member_repository: repositories::member::repository::Repository::new(),
      member_api: api::member::api::Api::new(api_config),
    }
  }

  pub async fn members(&self) -> Result<(), Box<dyn std::error::Error>> {
    let start_at = (chrono::Utc::now()).timestamp();

    if let Err(e) = async {
      let conn = match self.pool.get() {
        Ok(c) => c,
        Err(e) => {
          return Err(api::core::models::Error {
            code: -1,
            msg: Some(format!("pool.get error: {}", e)),
          });
        }
      };

      let req = api::member::models::ListMembersRequest { group_key: None };
      let res = match self.member_api.list_members(req).await {
        Ok(c) => c,
        Err(e) => {
          return Err(api::core::models::Error {
            code: -1,
            msg: Some(format!("member_api.list_members: {:?}", e)),
          });
        }
      };

      for u in res.users {
        if match self.member_repository.select(&conn, u.id) {
          Ok(c) => c,
          Err(e) => {
            return Err(api::core::models::Error {
              code: -1,
              msg: Some(format!("member_repository.select error: {}", e)),
            });
          }
        }
        .is_none()
        {
          let ss_get_member_by_username_req = bpr::ss::member::GetMemberByUsernameRequest {
            client: Some(bpr::models::core::network::Client {
              client_ip: self
                .connection_broker
                .client_ip()
                .unwrap_or(IpAddr::V4(Ipv4Addr::new(127, 0, 0, 1)))
                .to_string(),
              site_url: None,
              access_token: None,
            }),
            request: Some(bpr::ss::member::get_member_by_username_request::Request {
              username: u.site_username.clone(),
            }),
          };

          let ss_get_member_by_username_res_msg = match self
            .connection_broker
            .request(
              bpr::ss::member::SUBJECT_GET_MEMBER_BY_USERNAME,
              ss_get_member_by_username_req.encode_to_vec(),
            )
            .await
          {
            Ok(r) => r,
            Err(e) => {
              return Err(api::core::models::Error {
                code: -1,
                msg: Some(format!(
                  "bpr::ss::member::SUBJECT_GET_MEMBER_BY_USERNAME: {}",
                  e
                )),
              });
            }
          };

          let ss_get_member_by_username_res =
            match bpr::ss::member::GetMemberByUsernameResponse::decode(
              ss_get_member_by_username_res_msg.data.as_slice(),
            ) {
              Ok(r) => r,
              Err(e) => {
                return Err(api::core::models::Error {
                  code: -1,
                  msg: Some(format!(
                    "bpr::ss::member::GetMemberByUsernameResponse: {}",
                    e
                  )),
                });
              }
            };

          if let Some(e) = ss_get_member_by_username_res.error {
            return Err(api::core::models::Error {
              code: -1,
              msg: Some(format!(
                "bpr::ss::member::SUBJECT_GET_MEMBER_BY_USERNAME error: {}",
                e
              )),
            });
          }

          match ss_get_member_by_username_res.result {
            Some(r) => match r.member {
              Some(m) => {
                let member_id = uuid::Uuid::from_str(m.id.as_str()).expect("uuid::Uuid");

                match self.member_repository.insert(
                  &conn,
                  &repositories::member::models::NewMember {
                    id: u.id,
                    member_id,
                  },
                ) {
                  Ok(r) => r,
                  Err(e) => {
                    return Err(api::core::models::Error {
                      code: -1,
                      msg: Some(format!("member_repository.insert error: {}", e)),
                    });
                  }
                };
              }
              None => {
                return Err(api::core::models::Error {
                  code: -1,
                  msg: Some(format!("member is not exist: {}", u.site_username)),
                });
              }
            },
            None => {
              return Err(api::core::models::Error {
                code: -1,
                msg: Some(format!("member is not exist: {}", u.site_username)),
              });
            }
          };
        }

        self
          .member_repository
          .update(
            &conn,
            u.id,
            &repositories::member::models::ModifyMember {
              balance: u.cash,
              balance_bota: u.cash_bota,
              oriental_play: u.oriental_play,
            },
          )
          .expect("member update");
      }

      if let Err(e) = self.synchronization_history_repository.insert(
        &conn,
        &repositories::synchronization_history::models::NewSynchronizationHistory {
          item: repositories::synchronization::models::ITEM_MEMBERS.to_string(),
          start_at,
          complete_at: (chrono::Utc::now()).timestamp(),
          code: 0,
          message: None,
          data: None,
        },
      ) {
        println!("synchronization_history.insert error: {}", e);
      };

      Ok::<(), api::core::models::Error>(())
    }
    .await
    {
      let conn = match self.pool.get() {
        Ok(c) => c,
        Err(e) => {
          println!("pool.get error: {}", e);
          return Ok(());
        }
      };

      if let Err(e) = self.synchronization_history_repository.insert(
        &conn,
        &repositories::synchronization_history::models::NewSynchronizationHistory {
          item: repositories::synchronization::models::ITEM_MEMBERS.to_string(),
          start_at,
          complete_at: (chrono::Utc::now()).timestamp(),
          code: e.code,
          message: e.msg,
          data: None,
        },
      ) {
        println!("synchronization_history.insert error: {}", e);
      };
    }

    Ok(())
  }
}
